import datetime
import json

import yaml

from ..renamer.printer import PrinterInterface
from ..utils.jsons import ExtendedJsonEncoder
from ..utils.checksum import ChecksumOperator
from ..exceptions import CollectorError
from .storage import AssetStorage

# Non exhaustive list of Video containers with their file extension and name
# NOTE: May also describe some music only containers
MEDIAS_CONTAINERS = {
    "3gp": "3GPP",
    "asf": "Advanced Systems Format",
    "avi": "AVI",
    "flv": "Flash Video",
    "f4v": "Flash Video",
    "mov": "QuickTime",
    "mp4": "MPEG-4",
    "mkv": "Matroska",
    "mpg": "MPEG",
    "mpeg": "MPEG",
    "mpv": "MPEG",
    "mts": "MPEG Transport Stream",
    "qt": "QuickTime",
    "rm": "RealMedia",
    "ts": "MPEG Transport Stream",
    "vob": "Vob",
    "webm": "WebM",
    "wmv": "Windows Media Video",
}


# Default container label when extension does not match default container list
MEDIAS_DEFAULT_CONTAINER_NAME = "Unknow"


# List of unique file extensions for medias
MEDIAS_EXTENSIONS = set(MEDIAS_CONTAINERS.keys())


# Manifest filename to search in a directory
MANIFEST_FILENAME = "manifest.yaml"


# Forbidden/reserved keyword from manifest corresponding to computed values from
# collection (obviously excepted the ones from manifest)
MANIFEST_FORBIDDEN_VARS = {
    "path",
    "name",
    "absolute_dir",
    "relative_dir",
    "size",
    "mtime",
    "checksum",
    "children_files",
    "cover",
}


# File name to use with allowed extensions to search for a cover
COVER_NAME = "cover"


# Allowed file extensions to search for a cover. The order define the priority when
# there is multiple cover files in the same directory. The first extension will always
# have highest priority against other extensions.
COVER_EXTENSIONS = [
    ".png",
    ".jpg",
    ".jpeg",
    ".gif",
]


class Collector(PrinterInterface):
    """
    Collect informations about media files.

    Attributes:
        registry (dict): The registry where is collected all informations from scanning.
        stats (dict): Global statistics for all collected directories, files and total
            size.
        file_storage_queue (list): A list where each item is a tuple with source and
            destination to use for copying files.

    Arguments:
        basepath (pathlib.Path): The base directory for all directories to scan.
            Directories does not have to start directly from the basepath but must
            start from it.

    Keyword Arguments:
        extensions (list): A list of file extensions to consider as media files. A
            file extension that is not in this list are ignored. Default list contains
            all common video file extensions.
        allow_empty_dir (list): If True, even directory without direct media files are
            still collected in registry. Default is False, so only directories with
            direct media children are collected, the other directories are ignored from
            registry (but still scanned for children directories).
        manifest (string): Manifest filename to search for in a directory.
        cover_name (string): Cover file name (without extension) used to search for
            cover files.
        cover_extensions (list): Cover file extensions (with leading dot) used to
            search for cover files.
        allow_media_cover (boolean): If False, cover files will be ignored from dump.
            By default this is True and so covers are managed and dumped.
    """
    def __init__(self, basepath, extensions=MEDIAS_EXTENSIONS, allow_empty_dir=False,
                 manifest=MANIFEST_FILENAME, cover_name=COVER_NAME,
                 cover_extensions=COVER_EXTENSIONS, allow_media_cover=True):
        super().__init__()

        self.checksum_op = ChecksumOperator()
        self.basepath = basepath
        self.extensions = extensions
        self.allow_empty_dir = allow_empty_dir
        self.manifest_filename = manifest
        self.cover_name = cover_name
        self.cover_extensions = cover_extensions
        self.allow_media_cover = allow_media_cover
        self.file_storage_queue = []

        # Build elligible file names for cover from cover base file name and enabled
        # cover extensions
        self.cover_files = [
            self.cover_name + item
            for item in self.cover_extensions
        ]

        self.reset()

    def reset(self):
        """
        Reset registry and global states.

        This is the method to use if you plan to make multiple usage of ``run`` or
        ``scan_directory`` for different basepath since registry and global states are
        cumulative.
        """
        self.storage = AssetStorage(allowed_cover_filenames=self.cover_files)
        self.file_storage_queue = []

        self.registry = {}
        self.stats = {
            "directories": 0,
            "files": 0,
            "size": 0,
        }

    def timestamp_to_isoformat(self, timestamp):
        """
        Return datetime formatted from given timestamp.

        Arguments:
            timestamp (float): A timestamp as expected from date returned in
                ``Path.stat()``.

        Returns:
            string: Datetime formatted in ISO format without microseconds.
        """
        return datetime.datetime.fromtimestamp(
            timestamp,
            tz=datetime.timezone.utc
        ).isoformat(timespec="seconds")

    def _process_file_fields(self, fields, data):
        """
        Process field fields

        File field are collected as a tuple with file source path and destination path
        but only the destination path will be stored. The source path will just be
        used to copy the file source to its destination.

        Copying source file to destination is done through a queue to be performed
        after the end of collection.

        At this stage, we don't validate if a file item exist or not, since it has
        already be done during collection.

        Returns:
            dict: Given data possibly patched on file fields. Patch fields are
                transformed to just keep the final file path (not the original one).
        """
        for field in fields:
            if data.get(field):
                source, destination = data.get(field)
                self.file_storage_queue.append((source, destination))
                data[field] = destination

        return data

    def store(self, data):
        """
        Store given directory data.

        Arguments:
            data (dict): The data payload to store. It must have at least a ``path``
                item which will be used as the item key in the store.

        Returns:
            string: Item key name used to store the data.
        """
        key = str(data["path"].relative_to(self.basepath))

        self.registry[key] = self._process_file_fields(["cover"], data)

        return key

    def scan_file(self, path):
        """
        Scan a media file to get its informations.

        Implemention does not support file without any suffix (file extension).

        Arguments:
            path (pathlib.Path): File path to scan for informations.

        Returns:
            dict: Collected file informations.
        """
        # Get file stats informations
        stats = path.stat()

        relative_dir = path.parent.relative_to(self.basepath)

        dirname = path.parent.name
        # Prefer empty dirname instead of basepath dirname when file is at basepath
        # root
        if dirname == self.basepath.name:
            dirname = ""

        # Remove leading dot
        extension = path.suffix[1:].lower()
        # Get the media container label from file extension
        container = MEDIAS_DEFAULT_CONTAINER_NAME
        if extension in MEDIAS_CONTAINERS:
            container = MEDIAS_CONTAINERS[extension]

        data = {
            "path": path,
            "name": path.name,
            "absolute_dir": path.parents[0],
            "relative_dir": relative_dir,
            "directory": dirname,
            "extension": extension,
            "container": container,
            "size": stats.st_size,
            "mtime": self.timestamp_to_isoformat(stats.st_mtime),
        }

        self.stats["files"] += 1
        self.stats["size"] += data["size"]

        return data

    def get_directory_manifest(self, path):
        """
        Search for a YAML manifest to load medias informations related to
        given directory path.

        It should be safe to run with invalid manifests.

        Arguments:
            path (pathlib.Path): A Path object for the directory where to find
                manifest.

        Returns:
            dict: Manifest content.
        """
        manifest = {}
        manifest_path = path / self.manifest_filename

        if manifest_path.exists():
            try:
                manifest = yaml.load(manifest_path.read_text(), Loader=yaml.FullLoader)
            except yaml.YAMLError:
                msg = "No YAML object could be decoded for manifest: {}"
                self.log_warning(msg.format(str(manifest_path)))
            else:
                # Validate top level items against reserved keywords to avoid overriding
                # computed data from directory scan
                reserved = [
                    name
                    for name in MANIFEST_FORBIDDEN_VARS
                    if name in manifest
                ]
                if len(reserved) > 0:
                    msg = (
                        "Ignored manifest because it has forbidden keywords '{}': {}"
                    )
                    self.log_warning(msg.format(
                        ", ".join(reserved),
                        str(manifest_path),
                    ))
                    manifest = {}

        return manifest

    def scan_directory(self, path, checksum=False):
        """
        Scan a directory to get its media files.

        Arguments:
            path (pathlib.Path): Directory to scan for informations, for direct children
                files and to recursively search for children directories.

        Raises:
            CollectorError: If given path is not a directory inside
                basepath directory.

        Returns:
            dict: Directory information payload.
        """
        self.log_debug("Scanning {}".format(str(path)))

        try:
            relative_dir = path.relative_to(self.basepath)
        except ValueError:
            msg = "You cannot scan a directory which is out of given basepath: {}"
            raise CollectorError(msg.format(str(self.basepath)))

        # Get directory stats informations
        stats = path.stat()

        data = {
            "path": path,
            "name": path.name,
            "absolute_dir": path.parents[0],
            "relative_dir": relative_dir,
            "size": stats.st_size,
            "mtime": self.timestamp_to_isoformat(stats.st_mtime),
            "children_files": [],
        }

        for child in path.iterdir():
            if child.is_dir():
                self.scan_directory(child, checksum=checksum)
            else:
                if child.suffix and child.suffix.lower()[1:] in self.extensions:
                    data["children_files"].append(self.scan_file(child))

        # Only append directory datas if there is at least one file or empty dir is
        # allowed
        if self.allow_empty_dir or len(data["children_files"]) > 0:
            self.stats["directories"] += 1
            self.stats["size"] += data["size"]

            # Get possible manifest to extend data
            data.update(**self.get_directory_manifest(path))

            # Discover cover if any
            if self.allow_media_cover:
                data["cover"] = self.storage.get_directory_cover(path)

            # Perform content checksum if enabled
            if checksum:
                # Add file checksums
                self.checksum_op.payload_files(
                    data,
                    files_fields=["cover"],
                    storage=self.storage.storage_path,
                )
                # Then build directory info checksum
                data["checksum"] = self.checksum_op.directory_payload(
                    data,
                    files_fields=["cover"],
                    storage=self.storage.storage_path,
                )

            # Store collected data
            self.store(data)

        return data

    def run(self, destination=None, checksum=False):
        """
        Recursively scan everything from basepath to produce a registry of collected
        informations.

        Keyword Arguments:
            destination (pathlib.Path): Destination path to write a JSON file with
                every collected informations. Default is ``None`` so no JSON dump
                file will be written to the filesystem.
            checksum (boolean): Whether to enable directory checksums or not. Default
                to False, no checksum are done.

        Returns:
            dict: Dictionnary of global states for collected directories and files.
        """
        # Set storage basepath from destination location
        self.storage.set_basepath(destination, checksum=checksum)

        self.scan_directory(self.basepath, checksum=checksum)

        if self.registry and destination:
            with destination.open("w") as fp:
                json.dump(self.registry, fp, indent=4, cls=ExtendedJsonEncoder)
                self.log_info("Registry saved to: {}".format(str(destination)))

            # Proceed to copy queued files into storage dir
            self.storage.store_assets(self.file_storage_queue)

        return self.stats



import logging
from pathlib import Path

import click


try:
    import deepdiff  # NOQA: F401
    import tmdbv3api  # NOQA: F401
except ImportError:
    """
    Install is missing requirements from scrapping feature
    """
    @click.command()
    @click.pass_context
    def scrap_command(context):
        """
        The scrapping feature has not been installed and so this command is not
        available. See 'Install' documentation for details.
        """
        logger = logging.getLogger("deovi")

        logger.critical(
            "The scrapping feature has not been installed and so its command is not "
            "available. See 'Install' documentation for details."
        )

        raise click.Abort()
else:
    """
    Scrapping feature requirements are available
    """
    from ..scrapper import TmdbScrapper

    @click.command()
    @click.argument(
        "tvid",
        required=True,
    )
    @click.argument(
        "destination",
        type=click.Path(exists=False, path_type=Path),
    )
    @click.option(
        "--key",
        metavar="STRING",
        help=(
            "TMDb API private key as a string."
        ),
        default=None,
    )
    @click.option(
        "--filekey",
        type=click.Path(
            file_okay=True, dir_okay=False, resolve_path=False, path_type=Path,
            exists=True,
        ),
        help=(
            "TMDb API private key from a file, the file must only contain the key "
            "without anything else."
        ),
    )
    @click.option(
        "--language",
        metavar="STRING",
        help=(
            "Language code to get content."
        ),
        default="fr",
    )
    @click.option(
        "--write-diff",
        is_flag=True,
        help=(
            "If enabled, a file 'manifest.diff.txt' will be written along manifest "
            "file if there was a previous manifest file in destination directory and "
            "it got differences with the new one. This is not incremental, previous "
            "manifest difference file may be overwritten from a scrap job to another."
        ),
    )
    @click.pass_context
    def scrap_command(context, tvid, destination, key, filekey, language, write_diff):
        """
        Scrap TV show informations and poster image from TMDb API.

        Required arguments:

        TVID\n
            The TV Show ID from TMDb, it may looks like an integer ('14009').

        DESTINATION\n
            Destination directory path where to write manifest and cover files.
            If path does not exist it will be created.

        And finally a valid API Key is mandatory, give it either from option '--key'
        or '--filekey'.
        """
        logger = logging.getLogger("deovi")

        if not key and not filekey:
            logger.critical(
                "A TMDb API key is required either from 'key' or 'filekey' option."
            )
            raise click.Abort()
        elif filekey:
            key = filekey.read_text().strip()

        logger.info("TV ID: {}".format(tvid))
        logger.debug("Destination: {}".format(destination))
        logger.debug("Language: {}".format(language))
        logger.debug("API Key {}".format(
            "from file" if filekey else "from string",
        ))

        connector = TmdbScrapper(key, language=language)

        data, manifest, poster, diffs = connector.fetch_tv(
            destination,
            tvid,
            write_diff=write_diff,
        )
        logger.info("Title: {}".format(data["title"]))
        logger.info("Poster: {}".format(poster))
        if diffs:
            logger.info("There were differences with previous manifest file:")
            for line in diffs:
                logger.info("- {}".format(line))



import datetime
import shutil
import uuid
from pathlib import Path

from ..renamer.printer import PrinterInterface
from ..utils.checksum import ChecksumOperator


class AssetStorage(PrinterInterface):
    """
    Implement the asset storage logic.

    You can use the same instance for different basepaths but you will need to
    set each new base path with ``AssetStorage.set_basepath()``.

    Keyword Arguments:
        path (pathlib.Path): A directory or file path that will be used as the
            base path to store files. Note than asset files are not directly
            stored in the base path, they commonly are in their own subdirector
            from basepath. This argument can be empty but that will resolve
            basepath to the currently working directory which may not be very
            stable from code to another code.
        checksum (boolean): Whether to enable checksum or not. Default
            to False, asset storage paths won't any checksum included in their name.
        allowed_cover_filenames (list): List of filenames elligible as a directory
            cover file.
    """
    # Name used when given basepath is an empty Path
    DEFAULT_BASE_PATH = "attachment"

    def __init__(self, basepath=None, checksum=False, allowed_cover_filenames=None):
        super().__init__()

        self.checksum_op = ChecksumOperator()

        self.set_basepath(basepath, checksum=checksum)

        self.allowed_cover_filenames = allowed_cover_filenames or []

    def set_basepath(self, path=None, checksum=False):
        """
        Configure instance attributes for given base path.

        Keyword Arguments:
            path (pathlib.Path): A directory or file path that will be used as the
                base path to store files.
        """
        self.basepath = path
        self.storage_path = self.build_base_storage(self.basepath)
        self.storage_assets = self.build_assets_storage(
            self.basepath,
            checksum=checksum,
        )

    def build_base_storage(self, filepath):
        """
        Build base directory path.

        Nothing is writed on FS.

        Arguments:
            filepath (pathlib.Path):

        Returns:
            pathlib.Path: Either an empty Path or the filepath parent depending
            filepath is an empty Path or not.
        """
        if not filepath or str(filepath) == ".":
            return Path()

        return filepath.parent

    def build_assets_storage(self, filepath, checksum=False):
        """
        Build storage directory name from given filename and current datetime.

        Nothing is writed on FS.

        Arguments:
            filepath (pathlib.Path):

        Keyword Arguments:
            checksum (boolean): Whether to enable checksum or not. Default
                to False, asset storage paths won't any checksum included in its
                name.

        Returns:
            pathlib.Path: A filename composed from the filepath filename (without dirs
            or extension) and a computed unique hash.
        """
        if not filepath or str(filepath) == ".":
            filepath = Path(self.DEFAULT_BASE_PATH)

        if checksum:
            # Build hash from name + current ISO datetime
            suffix = self.checksum_op.filepath(filepath)
        else:
            # Build a simple datetime stamp
            suffix = datetime.datetime.now().isoformat(
                sep="T"
            ).replace(".", "").replace("-", "").replace(":", "")

        # Merge path stem with suffix
        return Path("{}_{}".format(filepath.stem, suffix))

    def get_directory_asset(self, path, filename_patterns):
        """
        Search for an asset file from given path.

        The first filename which match an allowed asset filename is returned. Order
        of ``filename_patterns`` defines matching order.

        Arguments:
            path (pathlib.Path): A Path object for the directory where to find
                cover image file.
            filename_patterns (list): A list of strings for asset filenames to search
                in directory.

        Returns:
            tuple: A tuple of two items ``(source, destination)`` where 'source' is the
                source cover file (Path object) resolved to an absolute path
                and 'destination' a filename (Path object) with a uuid4 instead of
                original source file name but with original source file extensions
                keeped.
        """
        for filename in filename_patterns:
            filepath = path / filename

            if filepath.exists():
                return (
                    filepath.resolve(),
                    self.storage_assets / Path(
                        "".join([str(uuid.uuid4()), filepath.suffix])
                    ),
                )

        return None

    def get_directory_cover(self, path):
        """
        Shortand around ``get_directory_asset`` to check for cover filenames.

        Arguments:
            path (pathlib.Path): A Path object for the directory where to find
                cover image file.

        Returns:
            tuple: A tuple with the format as from ``get_directory_asset`` returns.
        """
        return self.get_directory_asset(
            path,
            self.allowed_cover_filenames,
        )

    def store_assets(self, assets):
        """
        Store all given assets files into the assets directory.

        Assets are written to their destination path as given as second item of each
        asset, (first item is the source path).

        Arguments:
            assets (list): List of tuple ``(source, destination)`` where both items are
                Path objects as returned from ``Collector.get_directory_asset()``.

        Returns:
            list: List of stored file in their final destination.
        """
        stored = []

        if len(assets) > 0:
            container = self.storage_path / self.storage_assets

            if not container.exists():
                container.mkdir(parents=True, exist_ok=True)

            for source, destination in assets:
                if not source.exists():
                    msg = "File to store does not exists from your filesystem: {}"
                    self.log_warning(msg.format(source))

                # Destination path should be a relative path (from base) which already
                # include the assets directory
                shutil.copy(source, self.storage_path / destination)
                stored.append(self.storage_path / destination)

        return stored



import requests
import shutil
from pathlib import Path

from deepdiff import DeepDiff

from tmdbv3api import Configuration, TMDb, TV

import yaml


class TmdbScrapper:
    """
    Class to scrap informations from TMDb API.

    Arguments:
        api_key (string): Private key needed to use API.

    Keyword Arguments:
        language (string): Used language for payload content.
        poster_size (string): Size name as supported from TMDb API.
        poster_filename (string): Filename to use to write download poster image,
            without any extension.
    """
    def __init__(self, api_key, language="fr", poster_size="w780",
                 poster_filename="cover"):
        self.poster_size = poster_size
        self.poster_filename = poster_filename

        # Set TMDb client options
        self.client = self.get_client(api_key, language)

        # Get some config attributes from API
        self.get_api_configurations()

    def get_client(self, api_key, language):
        """
        Get client and set some options
        """
        # Set TMDb client
        tmdb = TMDb()
        tmdb.api_key = api_key
        tmdb.language = language
        tmdb.debug = True

        return tmdb

    def get_api_configurations(self):
        """
        Get configuration options returned by API

        This can only be done once the client has been initialized.
        """
        _api_configuration = Configuration()
        _api_infos = _api_configuration.info()

        # Entry point from TMDb API to download medias
        self.secure_base_url = _api_infos.images["secure_base_url"]

    def get_provider(self):
        """
        Shortcut to get the TV provider object from API client.
        """
        return TV()

    def get_poster_url(self, path):
        """
        Build URL to download poster image from API using its entrypoint and size
        """
        return "".join([
            self.secure_base_url,
            self.poster_size,
            path
        ])

    def serialize_tv_payload(self, provider, tv_id):
        """
        Get informations payload for given TV ID.
        """
        # Fetch payload from API
        payload = provider.details(tv_id)

        # print()
        # print("- id:", tv_id)
        # print("- name:", payload.name)
        # print("- status:", payload.status)
        # print("- episode_run_time:", payload.episode_run_time)
        # print("- first_air_date:", payload.first_air_date)
        # print("- number_of_episodes:", payload.number_of_episodes)
        # print("- number_of_seasons:", payload.number_of_seasons)
        # print("- poster_path:", payload.poster_path)
        # print("- genres:", [item["name"] for item in payload.genres])
        # print("- poster_url:", self.get_poster_url(payload.poster_path))

        return {
            "tmdb_id": tv_id,
            "tmdb_type": "tv",
            "title": payload.name,
            "status": payload.status,
            "poster_path": payload.poster_path,
            "first_air_date": payload.first_air_date,
            "number_of_seasons": payload.number_of_seasons,
            "number_of_episodes": payload.number_of_episodes,
            "genres": [item["name"] for item in payload.genres],
        }

    def fetch_poster(self, path, basepath):
        """
        Download poster from given url path and write it to basepath destination.
        """
        basefilepath = basepath / self.poster_filename

        url = self.get_poster_url(path)
        extension = Path(
            url.split("/")[-1]
        ).suffix

        destination = basefilepath.with_suffix(extension)

        # Go download the file
        with requests.get(url, stream=True) as r:
            # Create destination directory if missing
            if not basepath.exists():
                basepath.mkdir(parents=True, exist_ok=True)
            # Write file from stream
            with open(destination, "wb") as f:
                shutil.copyfileobj(r.raw, f)

        return destination

    def write_manifest(self, sourcepath, data, write_diff=False):
        """
        Write given data to manifest and possibly create a log file about differences
        with previous manifest file if any.
        """
        diff_lines = []

        # Write differences if any
        if sourcepath.exists():
            original = yaml.load(sourcepath.read_text(), Loader=yaml.FullLoader)
            diffs = DeepDiff(original, data)
            diff_lines = diffs.pretty().splitlines()
            if write_diff and diff_lines:
                diffpath = sourcepath.with_suffix(".diff.txt")
                diffpath.write_text("\n".join(diff_lines))

        # Write/overwrite manifest
        sourcepath.write_text(
            yaml.dump(data, Dumper=yaml.Dumper)
        )

        return diff_lines

    def fetch_tv(self, directory, tv_id, write_diff=False):
        """
        Get informations payload and medias for given TV ID.

        This downloads media files and build a YAML manifest to the given directory.
        """
        provider = self.get_provider()

        # Fetch and serialize TV show informations
        data = self.serialize_tv_payload(provider, tv_id)

        # Download possible poster image file in destination directory
        fetched_poster = None
        if data.get("poster_path", None):
            poster_path = data.pop("poster_path")
            fetched_poster = self.fetch_poster(poster_path, directory)

        # Build YAML manifest file to destination directory
        manifest = directory / "manifest.yaml"
        diff = self.write_manifest(manifest, data, write_diff=write_diff)

        return (
            data,
            manifest,
            fetched_poster,
            diff,
        )


if __name__ == "__main__":
    API_KEY = "c8a1f464dca620cde9037eda3f105425"
    connector = TmdbScrapper(API_KEY, language="fr")

    print()
    connector.fetch_tv(
        Path("/home/emencia/Projects/Apps/deovi/dist/StarTrek_Strange-New-Worlds"),
        "103516"
    )
    print()
    connector.fetch_tv(
        Path("/home/emencia/Projects/Apps/deovi/dist/Au dela du reel"),
        "21567"
    )



import datetime


DUMMY_ISO_DATETIME = "1977-06-02T07:35:15"


def timestamp_to_isoformat(*args, **kwargs):
    """
    A function which always returns the same string with datetime formatted to
    ISO format without microseconds.

    It is aimed to be used to mockup ``Collector.timestamp_to_isoformat`` in tests
    to patch file created/modified datetime since it is not stable through an
    installation to another, so we force it to a stable one.
    """
    return DUMMY_ISO_DATETIME


def dummy_uuid4():
    """
    A very dummy function to use to mockup ``uuid.uuid4()``.
    """
    return "dummy_uuid4"


class dummy_blake2b():
    """
    A class to mockup ``hashlib.blake2b()``.

    Instead of generating a hash it returns given content unchanged.

    Support basic string content, or 'update()' method and 'memoryview' so to be able
    to work with efficient memory/buffer way from 'checksum_file()'
    """
    def __init__(self, *args, **kwargs):
        self.content = args[0] if args else bytes()

    def update(self, content):
        if isinstance(content, memoryview):
            self.content += bytes(content)
        else:
            self.content += content

    def hexdigest(self, *args, **kwargs):
        """
        Arguments and keyword arguments are the same than original method.

        Returns:
            string: Given content, eventually decoded to string if it was a bytes.
        """
        # Convert to byte since hashlib do not accept string/unicode
        if isinstance(self.content, bytes):
            print("🐛 dummy_blake2b:", self.content, type(self.content))
            return self.content.decode("utf-8")

        return self.content


def dummy_checksum_file(cls, filepath):
    """
    A function to mockup ``checksum_file()``

    Instead of generating a hash from a file it just return the given file path.

    This may be required in tests which involves checksum_file with real file that
    will result to UnicodeDecodeError because of trying to decode binary as utf-8.

    Arguments:
        filepath (pathlib.Path): File path to open and checksum.

    Returns:
        string: The given file path.
    """
    print("🔖 dummy_checksum_file:", filepath, type(filepath))
    return str(filepath)


def dummy_checksumoperator_filepath(cls, filepath):
    """
    Support both ChecksumOperator.file and ChecksumOperator.filepath for monkey
    patching.

    Arguments and keyword arguments are the same than original methods.

    since the two
    receive a file Path object and will return path.

    Returns:
        string: Path from given file Path object.
    """
    content = "{}_{}".format(
        filepath.name,
        datetime.datetime.now().isoformat(),
    )

    print("✨ dummy_checksumoperator_filepath:", content, type(content))

    return content


def dummy_checksumoperator_directory_payload(cls, payload, files_fields=[],
                                             storage=None):
    """
    Support ``ChecksumOperator.directory_payload`` for monkey patching.

    Arguments and keyword arguments are the same than original method.

    Returns:
        string: JSON payload as done from ``ChecksumOperator.payload_files()``.
    """
    return cls.payload_files(
        payload,
        files_fields=files_fields,
        storage=storage
    )



import logging
from pathlib import Path

import click

from ..collector import MEDIAS_EXTENSIONS, Collector


@click.command()
@click.argument(
    "source",
    nargs=1,
    type=click.Path(exists=True, path_type=Path)
)
@click.argument(
    "destination",
    nargs=1,
    type=click.Path(exists=False, path_type=Path)
)
@click.option(
    "--extension",
    "-e",
    multiple=True,
    help=(
        "Give a specific file extension (without leading dot) to search for. You can "
        "use this argument multiple times for all extension you want to allow. On "
        "default the collector will use: {}".format(", ".join(MEDIAS_EXTENSIONS))
    ),
)
@click.option(
    "--checksum",
    is_flag=True,
    help=(
        "If enabled, collector will compute a checksum for each directory base on its "
        "collected informations, any directory change will have a different checksum "
        "from a dump to another."
    ),
)
@click.pass_context
def collect_command(context, source, destination, extension, checksum):
    """
    Recursively collect every directories with elligible media files from a basepath
    and dump it to a JSON file.

    The 'source' argument is a path which holds directories with media files to find and
    the 'destination' argument is a file path where to write the JSON dump.
    """
    logger = logging.getLogger("deovi")

    if not extension:
        extension = MEDIAS_EXTENSIONS

    logger.info("Source: {}".format(source))
    logger.info("Destination: {}".format(destination))
    logger.info("Extensions: {}".format(", ".join(extension)))

    collector = Collector(source, extensions=extension)

    stats = collector.run(destination=destination, checksum=checksum)

    logger.info("Registered directories: {}".format(stats["directories"]))
    logger.info("Registered files: {}".format(stats["files"]))
    logger.info("Total directories and files size: {}".format(stats["size"]))



from .collect import (
    MEDIAS_CONTAINERS, MEDIAS_DEFAULT_CONTAINER_NAME, MEDIAS_EXTENSIONS,
    MANIFEST_FILENAME, MANIFEST_FORBIDDEN_VARS, COVER_NAME,
    COVER_EXTENSIONS, Collector,
)
from .storage import AssetStorage


__all__ = [
    "MEDIAS_CONTAINERS",
    "MEDIAS_DEFAULT_CONTAINER_NAME",
    "MEDIAS_EXTENSIONS",
    "MANIFEST_FILENAME",
    "MANIFEST_FORBIDDEN_VARS",
    "COVER_NAME",
    "COVER_EXTENSIONS",
    "Collector",
    "AssetStorage",
]



import datetime
import json
import hashlib

from .jsons import ExtendedJsonEncoder


class ChecksumOperator:
    """
    Gather all methods which perform checksums.
    """
    def file(self, filepath):
        """
        Checksum a file in an efficient way for large files with blake2b.

        Borrowed from: https://stackoverflow.com/a/44873382

        Arguments:
            filepath (pathlib.Path): File path to open and checksum.

        Returns:
            string: The file checksum as 40 characters.
        """
        h = hashlib.blake2b()
        b = bytearray(128 * 1024)
        mv = memoryview(b)

        with open(filepath, "rb", buffering=0) as f:
            for n in iter(lambda: f.readinto(mv), 0):
                h.update(mv[:n])

        return h.hexdigest()

    def filepath(self, filepath):
        """
        Compute a string made up of filepath name and a blake2b checksum (build from
        filepath name with current datetime) divided by a ``_``.

        Arguments:
            filepath (pathlib.Path): Filepath used to compute a new name.

        Returns:
            string: New unique name.
        """
        return hashlib.blake2b(
            "{}_{}".format(
                filepath.name,
                datetime.datetime.now().isoformat(),
            ).encode("utf-8"),
            digest_size=10
        ).hexdigest()

    def payload_files(self, payload, files_fields=[], storage=None):
        """
        Patch directory payload to include file checksum.

        TODO: Payload should be patched to remove the file destination, since it may
              contains UUID4 and lead to directory checksum to always change, even
              directory has no changes. => NO, no removing destination here since
              collect expect it and also the checksum

        Arguments:
            payload (dict): The directory information payload to patch. Note than given
                dictionnary is mutated by the patch.

        Keyword Arguments:
            files_fields (list): A list of item names assumed to be file items to
                checksum.
            storage (pathlib.Path): A path to prefix all file paths if given. This is
                to use if you are storing relative paths (instead of absolute) in
                payload so a filepath can be resolved using this base storage path.

        Returns:
            dict: Patched payload.
        """
        # Before serialize, files_fields have to checksumed and payload altered with
        # their checksum
        for fieldname in files_fields:
            # Proceed on non empty field
            if payload.get(fieldname):
                # Get proper file path
                source, destination = payload[fieldname]
                if storage:
                    filepath = storage / source
                else:
                    filepath = source

                # Checksum the file if it exists
                key = "{}_checksum".format(fieldname)
                if filepath.exists():
                    payload[key] = self.file(filepath)
                else:
                    payload[key] = None

        return payload

    def directory_payload(self, payload_source, files_fields=[], storage=None):
        """
        Checksum directory informations including its files with blake2b.

        NOTE:
        File item are checksumed apart so they would change payload checksum if they
        change, we keep the file item untouched so if the file name change, this also
        triggers a payload checksum difference.

        Arguments:
            payload (dict): The directory information payload to checksum.

        Keyword Arguments:
            files_fields (list): A list of item names assumed to be file items to
                checksum.
            storage (pathlib.Path): A path to prefix all file paths if given. This is
                to use if you are storing relative paths (instead of absolute) in
                payload so a filepath can be resolved using this base storage path.

        Returns:
            string: The payload checksum as 40 characters.
        """
        # We do not want mutating of given payload
        payload = payload_source.copy()

        # Patch file item to only retain source path
        for fieldname in files_fields:
            if payload.get(fieldname):
                source, destination = payload[fieldname]
                payload[fieldname] = source

        # Serialize to JSON
        serialized = json.dumps(
            payload,
            indent=4,
            sort_keys=True,
            cls=ExtendedJsonEncoder
        )

        return hashlib.blake2b(serialized.encode("utf-8")).hexdigest()
